@@ -1,5 +1,5 @@
 /* mpc
- * (c) 2003-2004 by normalperson and Warren Dukes (shank@mercury.chem.pitt.edu)
+ * (c) 2003-2004 by normalperson and Warren Dukes (warren.dukes@gmail.com)
  *              and Daniel Brown (danb@cs.utexas.edu)
  * This project's homepage is: http://www.musicpd.org
  *
@@ -30,6 +30,7 @@
 #include <string.h>
 #include <unistd.h>
 #include <stdlib.h>
+#include <time.h>
 #include <sys/param.h>
 
 #define DIE(args...) do { fprintf(stderr, ##args); return -1; } while(0)
@@ -175,9 +176,52 @@ int cmd_add (int argc, char ** argv, mpd_Connection * conn )
 
 	/* clean up */
 	free(lists);
+	free(arglens);
 	return 0;
 }
 
+int cmd_crop( int argc, char ** argv, mpd_Connection * conn )
+{
+
+	mpd_Status *status;
+	status = getStatus( conn );
+
+	int length = ( status->playlistLength - 1 );
+
+	if( status->playlistLength == 0 ) {
+
+		mpd_freeStatus(status);
+		DIE( "You have to have a playlist longer than 1 song in length to crop" );
+
+	} else if( status->state == 3 || status->state == 2 ) { /* If playing or paused */
+
+		mpd_sendCommandListBegin( conn );
+		printErrorAndExit( conn );
+
+		while( length >= 0 )
+		{
+			if( length != status->song )
+			{
+				mpd_sendDeleteCommand( conn, length );
+				printErrorAndExit( conn );
+			}
+			length--;
+		}
+
+		mpd_sendCommandListEnd( conn );
+		my_finishCommand( conn );
+		mpd_freeStatus( status );
+		return ( 0 );
+
+	} else {
+
+		mpd_freeStatus(status);	
+		DIE( "You need to be playing to crop the playlist\n" );
+
+	}
+
+}
+
 int cmd_del ( int argc, char ** argv, mpd_Connection * conn )
 {
 	int i,j;
@@ -238,6 +282,77 @@ int cmd_del ( int argc, char ** argv, mpd_Connection * conn )
 	return 0;
 }
 
+int cmd_toggle( int argc, char ** argv, mpd_Connection * conn )
+{
+	mpd_Status * status;
+	status = getStatus(conn);
+
+	if(status->state==MPD_STATUS_STATE_PLAY) {
+		cmd_pause(0,0,conn);
+	} else {
+		cmd_play(0,0,conn);
+	}
+	return 1;
+}
+
+int cmd_outputs( int argc, char ** argv, mpd_Connection * conn )
+{
+	mpd_OutputEntity * output;
+
+	mpd_sendOutputsCommand( conn );
+	while(( output = mpd_getNextOutput( conn ))) {
+		/* We increment by 1 to make it natural to the user  */
+		output->id++;
+
+		/* If it's a negative number a password is needed  */
+		if( output->id > 0 )
+		{
+			if( output->enabled ) {
+				printf( "Output %i (%s) is enabled\n", output->id, output->name );
+			} else {
+				printf( "Output %i (%s) is disabled\n", output->id, output->name );
+			}
+		}
+		else
+		{
+			DIE( "error: cannot receive the current outputs\n" );
+		}
+		mpd_freeOutputElement( output );
+	}
+	mpd_finishCommand( conn );
+	return( 0 );
+}
+
+int cmd_enable( int argc, char ** argv, mpd_Connection * conn )
+{
+	int arg;
+
+        if( ! parse_int( argv[0], &arg ) || arg <= 0 ) {
+		DIE( "Not a positive integer\n" );
+	} else {
+		mpd_sendEnableOutputCommand( conn, (arg - 1) );
+	}
+
+	mpd_finishCommand( conn );
+	cmd_outputs( 0, 0, conn );
+	return( 0 );
+}
+
+int cmd_disable( int argc, char ** argv, mpd_Connection * conn )
+{
+	int arg;
+
+        if( ! parse_int( argv[0], &arg ) || arg <= 0 ) {
+		DIE( "Not a positive integer\n" );
+	} else {
+		mpd_sendDisableOutputCommand( conn, ( arg - 1 ) );
+	}
+	
+	mpd_finishCommand( conn );
+	cmd_outputs( 0, 0, conn );
+	return( 0 );
+}
+
 int cmd_play ( int argc, char ** argv, mpd_Connection * conn )
 {
 	int song;
@@ -260,21 +375,20 @@ int cmd_play ( int argc, char ** argv, mpd_Connection * conn )
 	return 1;
 }
 
-/* TODO: absolute seek times (normalperson) */
 int cmd_seek ( int argc, char ** argv, mpd_Connection * conn )
 {
 	mpd_Status * status;
 	char * arg = argv[0];
 	char * test;
-	int seekchange;
 	char * last_char;
-	int sec;
-	double perc;
+
+	int seekchange;
+	int total_secs;
 	int seekto;
         int rel = 0;
 
 	status = getStatus(conn);
-	
+
 	if(status->state==MPD_STATUS_STATE_STOP)
 		DIE("not currently playing\n");
 
@@ -285,26 +399,93 @@ int cmd_seek ( int argc, char ** argv, mpd_Connection * conn )
 
 	last_char = &arg[strlen(arg)-1];
 
-	if(*last_char == 's') {
-		/* absolute seek (in seconds) */
+	/* If seeking by percent */
+	if( *last_char == '%' ) {
 
-		sec = strtol(arg,&test,10); /* get the # of seconds */
+		double perc;
 
-		if(*test!='s' || (!rel && sec<0))
-			DIE("\"%s\" is not a positive number\n", arg);
-
-		seekchange = sec;
-
-	} else {
-		/* percent seek */
+		/* Remove the % */
+		arg[ strlen(arg) - 1 ] = '\0';
 
+		/* percent seek, strtod is needed for percent with decimals */
 		perc = strtod(arg,&test);
 
-		if(*test!='\0' || (!rel && (perc<0 || perc>100)) ||
-                                  (rel && perc>abs(100)))
+		if(( *test!='\0' ) || (!rel && (perc<0 || perc>100)) || (rel && perc>abs(100)))
 			DIE("\"%s\" is not an number between 0 and 100\n",arg);
 
 		seekchange = perc*status->totalTime/100+0.5;
+
+	} else { /* If seeking by absolute seek time */
+
+		if( strchr( arg, ':' )) {
+			char * sec_ptr;
+			char * min_ptr;
+			char * hr_ptr;
+
+			int hr = 0;
+			int min = 0;
+			int sec = 0;
+
+			/* Take the seconds off the end of arg */
+			sec_ptr = strrchr( arg, ':' );
+
+			/* Remove ':' and move the pointer one byte up */
+			* sec_ptr = '\0';
+			++sec_ptr;
+
+			/* If hour is in the argument, else just point to the arg */
+			if(( min_ptr = strrchr( arg, ':' ))) {
+
+				/* Remove ':' and move the pointer one byte up */
+				* min_ptr = '\0';
+				++min_ptr;
+
+				/* If the argument still exists, it's the hour  */
+				if( arg != NULL ) {
+					hr_ptr = arg;
+					hr = strtol( hr_ptr, &test, 10 );
+
+					if( *test != '\0' || ( ! rel && hr < 0 ))
+						DIE("\"%s\" is not a positive number\n", sec_ptr);
+				}
+			} else {
+				min_ptr = arg;
+			}
+
+			/* Change the pointers to a integer  */
+			sec = strtol( sec_ptr, &test, 10 );
+
+			if( *test != '\0' || ( ! rel && sec < 0 ))
+				DIE("\"%s\" is not a positive number\n", sec_ptr);
+
+			min = strtol( min_ptr, &test, 10 );
+
+			if( *test != '\0' || ( ! rel && min < 0 ))
+				DIE("\"%s\" is not a positive number\n", min_ptr);
+
+			/* If mins exist, check secs. If hrs exist, check mins  */
+			if( min && strlen(sec_ptr) != 2 )
+				DIE("\"%s\" is not two digits\n", sec_ptr);
+			else if( hr && strlen(min_ptr) != 2 )
+				DIE("\"%s\" is not two digits\n", min_ptr);
+
+			/* Finally, make sure they're not above 60 if higher unit exists */
+			if( min && sec > 60 )
+				DIE("\"%s\" is greater than 60\n", sec_ptr);
+			else if( hr && min > 60 )
+				DIE("\"%s\" is greater than 60\n", min_ptr);
+
+			total_secs = ( hr * 3600 ) + ( min * 60 ) + sec;
+
+		} else {
+
+			/* absolute seek (in seconds) */
+			total_secs = strtol( arg, &test, 10 ); /* get the # of seconds */
+
+			if( *test != '\0' || ( ! rel && total_secs < 0 ))
+				DIE("\"%s\" is not a positive number\n", arg);
+		}
+		seekchange = total_secs;
 	}
 
 	seekto = (rel ? status->elapsedTime + seekchange : seekchange);
@@ -801,6 +982,60 @@ int cmd_tab ( int argc, char ** argv, mpd_Connection * conn )
 	return 0;
 }
 
+char * DHMS(unsigned long t)
+{
+	static char buf[32];	/* Ugh */
+	int days, hours, mins, secs;
+
+#ifndef SECSPERDAY
+#define SECSPERDAY 86400
+#endif
+#ifndef SECSPERHOUR
+#define SECSPERHOUR 3600
+#endif
+#ifndef SECSPERMIN
+#define SECSPERMIN 60
+#endif
+
+	days = t / SECSPERDAY;
+	t %= SECSPERDAY;
+	hours = t / SECSPERHOUR;
+	t %= SECSPERHOUR;
+	mins = t / SECSPERMIN;
+	t %= SECSPERMIN;
+	secs = t;
+
+	snprintf(buf, sizeof(buf) - 1, "%d days, %d:%02d:%02d",
+	    days, hours, mins, secs);
+	return buf;
+}
+
+int cmd_stats ( int argc, char ** argv, mpd_Connection * conn )
+{
+	mpd_Stats *stats;
+	time_t t;
+
+	mpd_sendStatsCommand(conn);
+	stats = mpd_getStats(conn);
+	printErrorAndExit(conn);
+
+	if (stats != NULL) {
+		t = stats->dbUpdateTime;
+		printf("Artists: %6d\n", stats->numberOfArtists);
+		printf("Albums:  %6d\n", stats->numberOfAlbums);
+		printf("Songs:   %6d\n", stats->numberOfSongs);
+		printf("\n");
+		printf("Play Time:    %s\n", DHMS(stats->playTime));
+		printf("Uptime:       %s\n", DHMS(stats->uptime));
+		printf("DB Updated:   %s", ctime(&t));	/* no \n needed */
+		printf("DB Play Time: %s\n", DHMS(stats->dbPlayTime));
+		mpd_freeStats(stats);
+	} else {
+		printf("Error getting mpd stats\n");
+	}
+	return 0;
+}
+
 int cmd_status ( int argc, char ** argv, mpd_Connection * conn )
 {
 	return 1;
