@@ -1,5 +1,5 @@
 /* mpc
- * (c) 2003-2004 by normalperson and Warren Dukes (shank@mercury.chem.pitt.edu)
+ * (c) 2003-2005 by normalperson and Warren Dukes (warren.dukes@gmail.com)
  *              and Daniel Brown (danb@cs.utexas.edu)
  * This project's homepage is: http://www.musicpd.org
  *
@@ -71,7 +71,6 @@ void free_pipe_array (int max, char ** array)
 {
 	int i;
 	for ( i=0 ;i<max; ++i) free(array[i]);
-	free(array);
 }
 
 int get_boolean (const char * arg)
@@ -218,6 +217,7 @@ char * songToFormatedString (mpd_Song * song, const char * format, char ** last)
 	char * temp;
 	int length;
 	int found = 0;
+	int labelFound = 0;
 
 	/* we won't mess up format, we promise... */
 	for (p = (char *)format; *p != '\0'; )
@@ -294,39 +294,49 @@ char * songToFormatedString (mpd_Song * song, const char * format, char ** last)
 			end++;
 		}
 		length = end - p + 1;
+
+		labelFound = 0;
+		
 		if(*end != '%')
 			length--;
-		else if (strncmp("%file%", p, length) == 0)
+		else if (strncmp("%file%", p, length) == 0) {
 			temp = fromUtf8(song->file);
-		else if (strncmp("%artist%", p, length) == 0)
+		}
+		else if (strncmp("%artist%", p, length) == 0) {
+			labelFound = 1;
 			temp = song->artist ? fromUtf8(song->artist) : NULL;
-		else if (strncmp("%title%", p, length) == 0)
+		}
+		else if (strncmp("%title%", p, length) == 0) {
+			labelFound = 1;
 			temp = song->title ? fromUtf8(song->title) : NULL;
-		else if (strncmp("%album%", p, length) == 0)
+		}
+		else if (strncmp("%album%", p, length) == 0) {
+			labelFound = 1;
 			temp = song->album ? fromUtf8(song->album) : NULL;
-		else if (strncmp("%track%", p, length) == 0)
+		}
+		else if (strncmp("%track%", p, length) == 0) {
+			labelFound = 1;
 			temp = song->track ? fromUtf8(song->track) : NULL;
-		else if (strncmp("%name%", p, length) == 0)
+		}
+		else if (strncmp("%name%", p, length) == 0) {
+			labelFound = 1;
 			temp = song->name ? fromUtf8(song->name) : NULL;
-		else if (strncmp("%time%", p, length) == 0)
-		{
+		}
+		else if (strncmp("%time%", p, length) == 0) {
+			labelFound = 1;
 			if (song->time != MPD_SONG_NO_TIME) {
 				char s[10];
-				snprintf(s, 9, "%d:%d", song->time / 60, 
-						song->time % 60 + 1);
+				snprintf(s, 9, "%d:%02d", song->time / 60, 
+						song->time % 60);
 				/* nasty hack to use static buffer */
 				temp = fromUtf8(s);
 			}
 		}
 
-		if( temp == NULL)
-		{
-			/* just pass-through any unknown specifiers (including esc) */
-			/* drop a null char in so printf stops at the end of this specifier,
-			   but put the real character back in (pseudo-const) */
+		if( temp == NULL && !labelFound ) {
 			ret = appendToString(ret, p, length);
 		}
-		else {
+		else if( temp != NULL ) {
 			found = 1;
 			ret = appendToString(ret, temp, strlen(temp));
 		}
